{% extends "base.html" %}

{% block content %}
    <table>
        <thead>
        <tr>
            {% for column in headers %}
                <th>{{ column }}</th>
            {% endfor %}
        </tr>
        </thead>
        <tbody>
        {% for decoder in decoders %}
            <tr>
                {% for column in headers %}
                    <td>{{ decoder[column|lower] }}</td>
                {% endfor %}
            </tr>
        {% endfor %}
        </tbody>
    </table>
{% endblock %}